/* -*- Mode: C++ -*-
 * $Id: FuzzyFooDecision.cc,v 1.21 2001/07/31 11:15:24 sigge Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FuzzyFooDecision.cc
 * Unitname: Agent/FuzzyFoo
 * $Revision: 1.21 $
 * Created by: Mikael Brnnstrm, Andreas Bckert & Andreas Sigfridsson
 * Last modified by $Author: sigge $ $Date: 2001/07/31 11:15:24 $
 *
 *
 * HISTORY
 *
 * $Log: FuzzyFooDecision.cc,v $
 * Revision 1.21  2001/07/31 11:15:24  sigge
 * Defenders don't do passable, midfielders are also more conservative.
 *
 * Revision 1.20  2001/07/30 16:00:03  sigge
 * Player 1 is goalie.
 *
 * Revision 1.19  2001/07/30 11:26:27  mikbr212
 * nothing new really
 *
 * Revision 1.18  2001/07/29 14:43:09  sigge
 * Split decision tree into three parts.
 *
 * Revision 1.17  2001/07/29 11:40:19  sigge
 * Centralized intention debug.
 *
 * Revision 1.16  2001/07/28 14:33:11  sigge
 * Forwards now use Passable as their default intention.
 *
 * Revision 1.15  2001/07/28 13:00:01  mikbr212
 * Changed role
 *
 * Revision 1.14  2001/07/28 11:23:55  sigge
 * Removed fuzzyfoo.conf handling and modified decision tree quite a lot.
 *
 * Revision 1.13  2001/07/09 14:01:41  sigge
 * Better teleport and decision tree that uses the new intentions (Passable and AssistOffensive).
 *
 * Revision 1.12  2001/07/06 14:35:37  sigge
 * Fixed bug
 *
 * Revision 1.11  2001/07/06 14:12:53  sigge
 * FuzzyFooDecision now follows utility NWO
 *
 * Revision 1.10  2001/07/06 13:02:31  sigge
 * Removed IntentionDecision from various places.
 *
 * Revision 1.9  2001/07/05 12:45:48  sigge
 * Use utility for TakeBall
 *
 * Revision 1.8  2001/07/04 16:46:07  sigge
 * Teleport in appropriate game modes (and use dance formation before kick off)
 *
 * Revision 1.7  2001/07/04 13:21:36  sigge
 * Added role display.
 *
 * Revision 1.6  2001/07/03 13:27:32  andbo928
 * Added some intentions and uses the new weightet intention decision.
 *
 * Revision 1.5  2001/07/03 13:15:58  sigge
 * GoalieIntention now uses the old goalie behaviour. Old include directives removed.
 *
 * Revision 1.4  2001/07/03 12:46:53  sigge
 * Added ZoneDefence
 *
 * Revision 1.3  2001/06/19 16:37:59  mikbr212
 * Added comments
 *
 * Revision 1.2  2001/06/19 15:23:45  mikbr212
 * Debugged
 *
 * Revision 1.1  2001/06/19 14:30:24  mikbr212
 * Added team FuzzyFoo
 *
 *
 */

/**
 * @file FuzzyFooDecision.cc
 * Implementation of the class FuzzyFooDecision.
 *
 * @par Unit:
 *      Agent/FuzzyFoo
 * $Revision: 1.21 $
 * $Author: sigge $
 * $Date: 2001/07/31 11:15:24 $
 */

#include "FuzzyFooDecision.h"

// System headers
#include <unistd.h>
#include <errno.h>
#include <string>

// Views
#include "view_constants.h"
#include "CommunicationView.h"
#include "GameView.h"
#include "LookView.h"
#include "MarkerView.h"
#include "ParameterView.h"
#include "PassTreeView.h"
#include "Role.h"
#include "WorldStateView.h"

// Intentions
#include "intention_constants.h"
#include "AdvanceBallIntention.h"
#include "AssistOffensiveIntention.h"
#include "ContinueIntention.h"
#include "GoalieIntention.h"
#include "Intention.h"
#include "LookIntention.h"
#include "PassableIntention.h"
#include "ReceiveBallIntention.h"
#include "TakeBallIntention.h"
#include "ZoneDefenceIntention.h"

// Predicates
#include "BallKickedToMe.h"
#include "IAmClosestToBall.h"
#include "InPenaltyArea.h"
#include "PlayModePredicate.h"
#include "WeHaveBall.h"

// Framework
#include "ActuatorInterface.h"
#include "rs_assert.h"
#include "rs_log.h"

// Commands
#include "ChangeViewCommand.h"
#include "DashCommand.h"
#include "InitCommand.h"
#include "KickCommand.h"
#include "MoveCommand.h"
#include "SayCommand.h"
#include "TurnNeckCommand.h"
#include "TurnCommand.h"


RS_BEGIN_NAMESPACE

FuzzyFooDecision::FuzzyFooDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    parameterView(NULL),
    gameView(NULL),
    markerView(NULL),
    communicationView(NULL),
    worldStateView(NULL),
    lookView(NULL),
    passTreeView(NULL),
    serverAlive(true),
    busy(false),
    intentions((unsigned int)(MAX_INTENTION_ID+1), NULL)
#ifdef RS_USE_DISPLAY
  , theDisplay(Display::instance())
#endif
#ifdef RS_USE_COACH_DEBUGGING
  , coachView(NULL)
#endif
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(communicationView, COMMUNICATION_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(lookView, LOOK_VIEW_ID);

  getView(passTreeView, PASS_TREE_VIEW_ID);

#ifdef RS_USE_COACH_DEBUGGING
  getView(coachView, COACH_VIEW_ID);
#endif
  
  Role::init();
}


FuzzyFooDecision::~FuzzyFooDecision()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(COMMUNICATION_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(LOOK_VIEW_ID);
#ifdef RS_USE_COACH_DEBUGGING
  viewManager->releaseView(COACH_VIEW_ID);
#endif
}


ServerCommand* FuzzyFooDecision::initAgent()
{
  RS_REQUIRE( parameterView != NULL );
  InitCommand* init_cmd = new InitCommand(parameterView->CP_team_name, 5.5,
					  parameterView->CP_goalie);
  return init_cmd;
}


bool FuzzyFooDecision::start()
{
  RS_REQUIRE( initialized == true );
  while ( serverAlive )
  {
    if ( pause() == -1 && errno != EINTR )
      RS_ERROR("Something went wrong in pause!");
//     decide();
  }
  return true;
}

/** Choose an appropriate intention
 */
Intention* FuzzyFooDecision::chooseIntention() {
  static GoalieIntention goalie;
  static TakeBallIntention takeBall;   
  static ReceiveBallIntention receiveBall;
  static LookIntention look;       
  static AdvanceBallIntention advanceBall;
  static ZoneDefenceIntention zoneDefence;
  static PassableIntention passable;
  static AssistOffensiveIntention assistOffensive;

  static Role role;

  if(parameterView->CP_goalie || role.getRoleType() == Role::rtGoalie)
  {
    return &goalie;
  }
  else
  {
    // Common for all agents
    if(advanceBall.utility())
      return &advanceBall;
    if(takeBall.utility() >= Intention::uNormal)
      return &takeBall;

    if(look.utility() == Intention::uVeryGood)
      return &look;

    if(role.getRoleType() == Role::rtOffensive)
    {
      // *********
      // Offenders
      // *********

      if(passable.utility() >= Intention::uGood)
	return &passable;
      if(assistOffensive.utility() >= Intention::uGood)
	return &assistOffensive;

      if(look.utility() >= Intention::uGood)
	return &look;

      if(passable.utility() >= Intention::uNormal)
	return &passable;
      if(assistOffensive.utility() >= Intention::uNormal)
	return &assistOffensive;

      if(look.utility() >= Intention::uNormal)
	return &look;
      return &passable;
    }

    else if(role.getRoleType() == Role::rtDefensive)
    {
      // *********
      // Defenders
      // *********

      if(look.utility() >= Intention::uGood)
	return &look;

      if(zoneDefence.utility() >= Intention::uNormal)
	return &zoneDefence;

      if(passable.utility() >= Intention::uNormal)
	return &passable;

      if(look.utility() >= Intention::uNormal)
	return &look;
      return &zoneDefence;
    }

    else if(role.getRoleType() == Role::rtMid)
    {
      // *****************************
      // Flank defenders (midfielders)
      // *****************************

      if(passable.utility() >= Intention::uVeryGood)
	return &passable;
      if(assistOffensive.utility() >= Intention::uVeryGood)
	return &assistOffensive;

      if(look.utility() >= Intention::uGood)
	return &look;

      if(zoneDefence.utility() >= Intention::uNormal)
	return &zoneDefence;
      if(look.utility() >= Intention::uNormal)
	return &look;
      return &zoneDefence;
    }
  }
}

int lastCycle = -1;
bool FuzzyFooDecision::decide()
{
  if(busy)
    return true;
  busy = true;
  int gameTime = worldStateView->getWorldState().getGameTime();
  if(lastCycle == gameTime)
  {
    busy = false;
    return true;
  }
  lastCycle = gameTime;

  static Role role;
//   role.display();

  static Command* cmd = NULL;
  static Plan plan = Plan(Step(NULL));
  static Intention* chosenIntention = NULL;
  static Command* noCmd = new TurnCommand(0);
  
  const AgentState& agent(worldStateView->getWorldState().getAgent());

  if(communicationView->isThisMySpeakCycle(worldStateView->getWorldState()))
  {
    const WorldState& curWorld = worldStateView->getWorldState();
    string finstreng = worldStateView->getWorldState().serialize();
    communicationView->addMessage(Message(worldStateMessageType, finstreng));
    string message = communicationView->getOutgoing();
    if(message.size() > 510)
      cout << "Too large message!" << endl;
    else
    {
      SayCommand sayCommand(message);
      actuatorInterface->addToPlan(&sayCommand);
    }
  }

  // ** If the plan is empty, then pick an intention and generate a plan **
  vector<PlayMode> playModes = gameView->getPlayModes(NOW);
  if(find(playModes.begin(), playModes.end(), PM_BEFORE_KICK_OFF) != playModes.end()
     || find(playModes.begin(), playModes.end(), PM_HALF_TIME) != playModes.end()
     || find(playModes.begin(), playModes.end(), PM_THEIR_GOAL) != playModes.end()
     || find(playModes.begin(), playModes.end(), PM_OUR_GOAL) != playModes.end())
  {
    TVector offset(0, 0);
    if(find(playModes.begin(), playModes.end(), PM_BEFORE_KICK_OFF) != playModes.end())
    {
      // Dance mode
//       int time = worldStateView->getWorldState().getGameTime();
//       float angle = 360.0*float(time % 40)/40.0;
//       offset = 5.0*TVector(cos(angle), sin(angle));
    }
    TVector position = role.getInitPosition() + offset;
    if(position.getX() > -1)
      position.setX(-1);
    Point p(position.getX(), position.getY());
    MoveCommand move_cmd(p);
    actuatorInterface->addToPlan(&move_cmd);
  }

  if(plan.empty()) {
    // choose intention
    chosenIntention = chooseIntention();
    if(parameterView->CP_debug)
    {
      string s;
      switch(chosenIntention->getID())
      {
      case TAKE_BALL_INTENTION_ID:		s = "T"; break;
      case ADVANCE_BALL_INTENTION_ID:		s = "A"; break;
      case ZONE_DEFENCE_INTENTION_ID:		s = "Z"; break;
      case PASSABLE_INTENTION_ID:		s = "P"; break;
      case ASSIST_OFFENSIVE_INTENTION_ID:	s = "a"; break;
      case LOOK_INTENTION_ID:			s = "L"; break;
      case GOALIE_INTENTION_ID:			s = "G"; break;
      case RECEIVE_BALL_INTENTION_ID:		s = "R"; break;
      default:
	s = "U";
      }
      cerr << s << flush;
    }
    worldStateView->setIntention(chosenIntention);
    
    // apply intention
    if(chosenIntention != NULL) 
      plan = chosenIntention->generatePlan();

  } // end if(plan.empty())
  
  // ** Evaluate the first step in the plan **
  if(!plan.empty()) {
    Step step = plan.getNextStep();
    plan.removeNextStep();
    while(!step.empty()) {
      cmd = step.getNextCommand();
      step.removeNextCommand();
      if ( cmd != NULL ) {
	actuatorInterface->addToPlan(cmd);
	delete cmd;
      }
    } 
  } else {
    actuatorInterface->addToPlan(noCmd); // avoid spam
  }

  busy = false;
  return true;
  
}


void FuzzyFooDecision::onActuatorSensorData()
{
#ifdef RS_USE_DISPLAY
//    theDisplay->setLastSkill(commandView->getLastServerString());
#endif
#ifdef RS_USE_OWN_LOGFILE
//    logFile << gameView->getCurrentTime() << ": executed action "
//  	  << commandView->getLastServerString()
//  	  << std::endl;
//    if ( commandView->getActualTurn(NOW) > AngleDeg(0) )
//      logFile << gameView->getCurrentTime() << ": actual turn = "
//  	    << commandView->getActualTurn(NOW)
//  	    << std::endl;  
#endif
}


void FuzzyFooDecision::onInit()
{
  // Let the goalie use narrow vision to get rapid updates
  if(parameterView->CP_goalie) {
    ChangeViewCommand cvc(VW_NARROW, VQ_HIGH);
    actuatorInterface->addToPlan(&cvc);
  } else {
//      ChangeViewCommand cvc(VW_NARROW, VQ_HIGH);
    ChangeViewCommand cvc(VW_NORMAL, VQ_HIGH);
    actuatorInterface->addToPlan(&cvc);
  } 

  const AgentState& agent(worldStateView->getWorldState().getAgent());

  UInt16 no = agent.getUniformNumber();
  if(no < 1 || no > 11)
    no = 1;
  if(parameterView->CP_goalie && no != 1)
  {
    cerr << "*******************************************" << endl
	 << "Player is goalie but does not have number 1" << endl
	 << "*******************************************" << endl;
  }
}


void FuzzyFooDecision::onVisualSensorData()
{
  //RS_LOG( LA_DECISION, gameView->getCurrentTime() << ": **saw**" );
}


void FuzzyFooDecision::onAfterSensorData()
{
#ifdef RS_USE_DISPLAY
  theDisplay->draw();
#endif
}


void FuzzyFooDecision::onNewCycle(const bool forced)
{
  /*  if ( forced )
      RS_LOG( LA_DECISION,
      gameView->getCurrentTime() << ": **Forced NewCycle**" );
      else
      RS_LOG( LA_DECISION, gameView->getCurrentTime() << ": **NewCycle**" );
  */
}


bool FuzzyFooDecision::onEstimatedNewCycle(UInt16 estimated_time, UInt16 ticks)
{
  // We assume that if the sense_body is more than four ticks late
  // then it's lost.
  if ( false )
    std::cout << "onEstimatedNewCycle(" << estimated_time
	      << ", " << ticks << ")" << endl;

  // If more than nine sense_body's are missed in a row then assume that
  // the server is dead.
  if ( estimated_time > 9 )
    serverAlive = false;

  if ( ticks > 4 ) {
    // TODO: change this into sending a control command instead!
    //sensorInterface->ForceNewCycle();
    return true;
  } else
    return false;
}


void FuzzyFooDecision::onCommandWarning()
{
  decide();
}


RS_END_NAMESPACE

