/* -*- Mode: C++ -*-
 * $Id: Intention.h,v 1.7 2001/07/06 14:16:31 mikbr212 Exp $
 *
 * IDENTIFICATION
 *
 * $Revision: 1.7 $
 * Created by: Mikael Brnnstrm, Andreas Bckert, Andreas Sigfridsson
 * Last modified by $Author: mikbr212 $ $Date: 2001/07/06 14:16:31 $
 *
 *
 * HISTORY
 *
 * $Log: Intention.h,v $
 * Revision 1.7  2001/07/06 14:16:31  mikbr212
 * sucks -> suck
 *
 * Revision 1.6  2001/07/06 14:05:24  mikbr212
 * changed method utility()
 *
 * Revision 1.5  2001/07/06 12:53:37  mikbr212
 * Changed utility-methods
 *
 * Revision 1.4  2001/07/06 12:30:36  mikbr212
 * Added descriptions of utility
 *
 * Revision 1.3  2001/07/06 10:00:30  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 * Revision 1.2  2001/07/03 15:24:54  andbo928
 * Added utility methods and enum.
 *
 * Revision 1.1  2001/06/19 14:11:43  mikbr212
 * Created
 *
 */

/**
 * @file Intention.h
 * Definition file for the class Intention.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.7 $
 * $Author: mikbr212 $
 * $Date: 2001/07/06 14:16:31 $
 */

#ifndef INTENTION_H
#define INTENTION_H

#include "intention_constants.h"
#include "view_constants.h"
#include "ViewManager.h"
#include "FuzzyBool.h"
#include "Plan.h"
#include "Skill.h"
#include "PlayerState.h"

#include "WorldStateView.h"
#include "PassTreeView.h"
#include "StatisticsView.h"

RS_BEGIN_NAMESPACE

class ParameterView;
class GameView;
//  class StatisticsView;

//class Skill; // To avoid circular references.

/** 
 * @class Intention Intention.h
 * This class is the abstract base class for all intentions.
 *
 * $Revision: 1.7 $
 * $Author: mikbr212 $
 * $Date: 2001/07/06 14:16:31 $
 */
class Intention {
public:

  /** Enum describing the different utility values of an intention.
   * <br><ul>
   * <li><b>uSuck:</b> Don't call this intention. Not applicable. 
   * The value of uSucks is zero, so it can be used as a boolean.</li>
   * <li><b>uVeryBad:</b> Lousy result/insecure effect, call me only if
   * nothing else is better.</li>
   * <li><b>uBad:</b> Probably not good.</li>
   * <li><b>uNormal:</b> The result might be good. Not sure.</li>
   * <li><b>uGood:</b> Probably good, a little more confident than uNormal.</li>
   * <li><b>uVeryGood:</b> The result will be very good for sure.</li>
   * </ul>
   */
  enum Utility
  {
    uSuck = 0,
    uVeryBad = 1,
    uBad = 2,
    uNormal = 3,
    uGood = 4,
    uVeryGood = 5
  };
  

  /** Constructs an intention with default behavior. */
  Intention();

  /** Destroys the intention. */
  virtual ~Intention();

  /** Generate a plan for how to accomplish the goal
   * of the intention. This is often done by choosing
   * a skill and returning the plan which was generated by the skill.
   * The default behaviour is therefore to return getPlan(false).
   * Either overload this method or chooseSkill() to implement an
   * intention of your own.
   * @see getPlan(bool), chooseSkill()
   */
  virtual Plan generatePlan();

  /** Return @c the utility value of the intention.
   * The agent is considered.
   */
  virtual Utility utility() = 0;

  /** Return @c the utility value of the intention.
   * This should be a light weight method, since this
   * will be called for all other players.
   */
  virtual Utility modelUtility(const WorldState&) = 0;

  /** Models a player. This method determines which action(s)
   * should be done by the player according to this intention.<br>
   * This method should not be called
   * for the agent, the framework (WorldStateView) will
   * compensate for the actions done by the agent. 
   */
  virtual void model(WorldState&) = 0;

  /** Returns a float between 0 and 1, which tells how
   * persistent this intention is. This value should
   * be a static value that doesn't change over time, instead
   * it should represent the nature of this skill. ie is
   * the intention a short time intention (close to zero)
   * or is the intention a long time intention that wants to
   * be called several steps in a row (close to one).
   */
  virtual float getPersistence() const = 0;

  /** Returns the ID that identifies this (type of) intention.
   */
  virtual IntentionId getID() const = 0;
  
protected:
  /** Chooses a skill. It works like a mini-decision maker.
   * Since this is an abstract class the implementation
   * of this method simply returns NULL.
   */
  virtual Skill* chooseSkill();

  /** Chooses a skill if the last chosen skill is not
   * persistent and applicable. Then it returns a plan
   * from that skill.
   * @param clear if true, then the last chosen skill is
   * cleared and a new skill be chosen. Otherwise the
   * persistence and applicability of the last skill
   * will be checked (default).
   */
  virtual Plan getPlan(bool clear=false);

  /** A pointer to the view manager. */
  ViewManager* const viewManager;

  /** A pointer to the parameter view. */
  ParameterView* parameterView;

  /** A pointer to the world state view. */
  WorldStateView* worldStateView;

  /** A pointer to the pass tree view. */
  PassTreeView* passTreeView;

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the game view. */
  StatisticsView* statisticsView;

private:
  /** Hide copy-constructor. */
  Intention(const Intention&);
  /** Hide assignment operator. */
  Intention& operator= (const Intention&);
};

RS_END_NAMESPACE

#endif // INTENTION_H
