
/* -*- Mode: C++ -*-
 * $Id: FuzzyScore.cc,v 1.3 2001/07/01 12:34:45 andbo928 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FuzzyScore.cc
 * Unitname: Framework/skill
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: andbo928 $ $Date: 2001/07/01 12:34:45 $
 *
 *
 * HISTORY
 */

/**
 * @file FuzzyScore.cc
 * Implementation file for the skill class FuzzyScore.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: andbo928 $
 * $Date: 2001/07/01 12:34:45 $
 */


#include "FuzzyScore.h"

RS_BEGIN_NAMESPACE


FuzzyScore::FuzzyScore()
  : Skill(),
    //    kickSkill(Vector()),
    parameterView(NULL),
    markerView(NULL),
    worldStateView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
}


FuzzyScore::~FuzzyScore()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}

/**
 * Calculate best goalvector. Relative length. Absolute angle.
 */
TVector FuzzyScore::calculateVector(float& good) {
  const WorldState& worldState = worldStateView->getWorldState();
  const TVector& pos = worldState.getAgent().getPosition();
  TVector toTheirGoal = TVector(markerView->getTheirGoal(NOW).getPosition()) - pos;

  double goalWidth = parameterView->SP_goal_width*0.92;
  TVector goalL = toTheirGoal - TVector(0, goalWidth/2);
  TVector goalR = toTheirGoal + TVector(0, goalWidth/2);
  double leftAng = goalL.getAngle();
  double rightAng = goalR.getAngle();
  // leftAng will be smaller than rightAng

  TLine goalLine(goalL, goalR);

  WorldState WSNextCycle = worldStateView->getWorldState();
  WSNextCycle.step();
  std::vector<PlayerState> ops = WSNextCycle.getOpponents();
  std::vector<PlayerState>::iterator it;
  std::vector<double> threats;
  threats.push_back(leftAng);
  threats.push_back(rightAng);

  TVector opPos;
  for(it = ops.begin(); it != ops.end(); it++) {
    if(it->getC() > 0.1) {
      opPos = it->getPosition() - pos;
      if(opPos.getAngle() > leftAng &&
	 opPos.getAngle() < rightAng) 
	threats.push_back(opPos.getAngle());
    }
  } // for

  // Sort the angles!
  // Bubbles are beutiful!

  bool sorted = false;
  double tmp;
  while(!sorted) {
    sorted = true;
    for(unsigned int i = 0; i < threats.size()-1; i++) {
      if(threats[i] > threats[i+1]) {
	tmp = threats[i];
	threats[i] = threats[i+1];
	threats[i+1] = tmp;
	sorted = false;
      }
    }
  }

  float maxAngle = 0;
  float angle;
  TVector bestVec(1, 0);
  for(unsigned int i = 0; i < threats.size()-1; i++) {
    angle = threats[i+1] - threats[i];
    //        if((i == 0 || 
    //  	  i == threats.size()-2) &&
    //  	 (threats.size() > 2))
    //  	angle *= 1.9;
      
    if(maxAngle < angle) {
      maxAngle = angle;
      if(threats.size() == 2)
	bestVec.setAngle(threats[i] + angle/2);
      else if(i == 0)
	bestVec = goalLine.getStart() + TVector(0, 3);
      else if(i == threats.size()-2)
	bestVec = goalLine.getEnd() - TVector(0, 3);
      else
	bestVec.setAngle(threats[i] + angle/2);
    }
  }
  bestVec = TLine(TVector(0, 0), bestVec).getIntersection(goalLine);
  const float scale = (1-parameterView->SP_ball_decay)/
    (1 - pow(parameterView->SP_ball_decay,16));
  bestVec *= scale;

  if(bestVec.getLength() < 1.3)
    bestVec.setLength(1.3);

  good = maxAngle;

  return bestVec;
}

Plan FuzzyScore::generatePlan()
{
  float f;
  TVector kickVec = calculateVector(f);
  kickVec.setLength(parameterView->SP_ball_speed_max);
  WorldState worldState = worldStateView->getWorldState();

  worldState.doKickToSpeed(kickVec);
  Plan p(worldState.getStep());
  return p;

//    if(kickVec.getLength() > 1.31)
//      kickVec.setLength(3.0);
  //  kickSkill.setVector(kickVec);
  // return kickSkill.generatePlan();
}

FuzzyBool FuzzyScore::persistent()
{
  return kickSkill.persistent();
}

FuzzyBool FuzzyScore::applicable()
{
  const WorldState& worldState = worldStateView->getWorldState();
  if(worldState.getBall().getC() > 0.6
     && worldState.getAgent().isBallKickable(worldState.getBall()))
  {
      float angle;
      TVector kickVec = calculateVector(angle);
      //      cout << "kickVec = " << kickVec.getLength() << endl;;
      
      if(kickVec.getLength() < 2.5)
	return FuzzyBool(/*min(*/(double)angle/45/*, 
			     1.0/kickVec.getLength())*/);
  }
  return false;
}


RS_END_NAMESPACE

