/**
 * @file FuzzyTeleport.cc
 * Implementation of the FuzzyTeleport skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.6 $
 * $Author: sigge $
 * $Date: 2001/04/24 21:55:37 $
 */

#include "FuzzyTeleport.h"
#include "MoveCommand.h"
#include "TVector.h"
#include <algorithm>

#define KICK_RANGE 30
#define LOOK_RANGE 15
#define SAFE_MARGIN 2
#define CB_FRIENDLY 0.2
#define CB_OPPONENT -1.0
#define CB_ADVANCE  0.5
#define CB_PRECISION 0.0

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

FuzzyTeleport::FuzzyTeleport()
  : Skill(),
    parameterView(NULL), //,
    markerView(NULL),
    gameView(NULL),
    worldStateView(NULL)
{
  lastTeleport = 0;
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(gameView, GAME_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
}

// --- DESTRUCTOR ---

FuzzyTeleport::~FuzzyTeleport()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}


// --- METHODS ---

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

Plan FuzzyTeleport::generatePlan()
{
  lastTeleport = worldStateView->getWorldState().getGameTime();
  Point movePos = calculateMovePos();
  MoveCommand cmd(movePos);
  return Plan(Step(&cmd));
}

Point FuzzyTeleport::calculateMovePos() {

  // Construct the edge lines.


  TVector topFlag = TVector(markerView->getMarker(FLAG_PLT, NOW).getPosition())
    - TVector(0.1, 0);
  TVector botFlag = TVector(markerView->getMarker(FLAG_PLB, NOW).getPosition())
    - TVector(0.1, 0);

  /*  AngleDeg dir = territoryControlView->getMaxMinNoCache(CB_FRIENDLY, 
							CB_OPPONENT, 
							CB_ADVANCE, 
							CB_PRECISION,
							clearSafeness, 
							topV.getAngle(),
							botV.getAngle());

//   cout << "Cleardir: " << dir << endl;

  if(agentView->getBodyDirection(NOW).isKnown()) {
    dir = dir - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle();
  }
  */

  Float unsafeDirs[128];
  int numUnsafeDirs = 0;

  //  cout << "limits: " << minAng << ", " << maxAng << endl;
  
  unsafeDirs[numUnsafeDirs] = topFlag.getY();
  numUnsafeDirs++;
  unsafeDirs[numUnsafeDirs] = botFlag.getY();
  numUnsafeDirs++;

  const WorldState worldState = worldStateView->getWorldState();
  const std::vector<PlayerState> oppVec = worldState.getOpponents();

  for(unsigned int i = 0; i < oppVec.size(); i++) {
    if(oppVec[i].getC() > 0.05 &&
       oppVec[i].getPosition().getX() < LOOK_RANGE) {
      TVector plPnt = oppVec[i].getPosition();
      if(plPnt.getY() > topFlag.getY() && plPnt.getY() < botFlag.getY()) {
	if(plPnt.getX() > topFlag.getX()) {
	  unsafeDirs[numUnsafeDirs] = plPnt.getY();
	  numUnsafeDirs++;
	}
      }
    }
  }


//   oppVec = playerView->getUnknownPlayers(NOW);

//   for(unsigned int i = 0; i < oppVec.size(); i++) {
//     if(oppVec[i].getPos().getX().getConfidence() > 0.05 &&
//        oppVec[i].getPos().getX() < LOOK_RANGE) {
//       Point plPnt = oppVec[i].getPos().getPoint();
//       if(plPnt.getY() > topFlag.getY() && plPnt.getY() < botFlag.getY()) {
// 	if(plPnt.getX() > topFlag.getX()) {
// 	  unsafeDirs[numUnsafeDirs] = plPnt.getY();
// 	  numUnsafeDirs++;
// 	}
//       }
//     }
//   }

  // Sort the dirs!
  // Bubbles are beutiful!

  bool sorted = false;
  Float tmp;
  while(!sorted) {
    sorted = true;    for(int i = 0; i < numUnsafeDirs-1; i++) {
      if(unsafeDirs[i] > unsafeDirs[i+1]) {
	tmp = unsafeDirs[i];
	unsafeDirs[i] = unsafeDirs[i+1];
	unsafeDirs[i+1] = tmp;
	sorted = false;
      }
    }
  }

  /*  cout << "Unsafe: [";
  for(int i = 0; i < numUnsafeDirs; i++)
    cout << unsafeDirs[i] << ",";

    cout << "]" << endl; */

  // Calc the largest interval.

  Float maxStart = 0, maxEnd, maxVal;
  maxVal = -1;

  for(int i = 0; i < numUnsafeDirs-1; i++) {
    Float curVal = unsafeDirs[i+1] - unsafeDirs[i];
    if(curVal > maxVal) {
      maxStart = unsafeDirs[i];
      maxEnd = unsafeDirs[i+1];
      maxVal = curVal;
    }
  }

  // WhoppidooooO!

  //  cout << "Wanted dir" << (maxStart + maxVal / 2 ) << endl;

  Point pnt = Point(topFlag.getX(), maxStart + maxVal / 2);
  
  return pnt;
}


/**
 * Persistent.
 * FALSE
 */
FuzzyBool FuzzyTeleport::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value
 */

FuzzyBool FuzzyTeleport::applicable()
{
  const WorldState worldState = worldStateView->getWorldState();
  if(!worldState.getAgent().isBallKickable(worldState.getBall()))
    return FALSE;

  // Did we teleport last few cycles?
  if(lastTeleport >= worldState.getGameTime() - 3)
    return FALSE;

  std::vector<PlayMode> playModes = gameView->getPlayModes(NOW);
  if(find(playModes.begin(), playModes.end(), PM_OUR_FREE_KICK) != playModes.end())
    return TRUE;

  return FALSE;
}

RS_END_NAMESPACE
