/**
 * @file MarkPosition.cc
 * Implementation of the MarkPosition skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.11 $
 * $Author: mikbr212 $
 * $Date: 2001/08/01 14:17:00 $
 */

#include "Line.h"
#include "MarkPosition.h"
#include "TAngle.h"
#include "StatisticsView.h"
#include "Display.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

MarkPosition::MarkPosition()
  : Skill(),
    worldView(NULL),
    statisticsView(NULL),
    markLine(TVector(), TVector()),
    moveLine(TVector(), TVector())
{
  getView(worldView, WORLD_STATE_VIEW_ID);
  getView(statisticsView, STATISTICS_VIEW_ID);

  // Fix defaults
  setLookAtMark();
  setBuffer();
  setDistance();
  setBallAsDistantPosition();

  moveSkill.setLookAtBall(true);
  moveSkill.setSpeed(MoveToPosition::msFast);
}

// --- DESTRUCTOR ---
MarkPosition::~MarkPosition()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}

// --- METHODS ---
void MarkPosition::setPosition(const TVector& pos)
{
  position = pos;
}

void MarkPosition::setLookAtMark(bool enabled)
{
  lookAtMark = enabled;
}

void MarkPosition::setBuffer(double bufSize)
{
  distanceBuffer = bufSize;
}

void MarkPosition::setDistance(double dist)
{
  distance = dist;
}

void MarkPosition::setDistantPosition(const TVector& pos) {
  distantPosition = pos;
  ballIsDistantPosition = false;
}

void MarkPosition::setBallAsDistantPosition(bool flag) {
  ballIsDistantPosition = flag;
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */
Plan MarkPosition::generatePlan()
{
  WorldState ws = worldView->getWorldState();

  // Construct the markLine
  if(ballIsDistantPosition)
  {
    // Use threat forwarding:
    // If the ball is passed to another player, use the predicted
    // interception point as the threat
    const BallState& threat = statisticsView->getInterceptionBallState();
    Display::instance()->addArrow(position, threat.getPosition());
    markLine = TLine(position, threat.getPosition());
  }
  else
    markLine = TLine(position, distantPosition);

  // Construct the moveLine
  moveLine = TLine(ws.getAgent().getPosition(),
		   ws.getAgent().getPosition() 
		   + ws.getAgent().getDirectionVector());

  // Calculate the next marking position.
  TVector nextPos;
  if(markLine.isParallel(moveLine))
  {
    nextPos = TVector(-666, -666);
  }
  else
  {
    nextPos = markLine.getIntersection(moveLine);
  }

  const int LK_Distant=0;
  const int LK_Mark=1;

  static int lastLook = LK_Mark;

  // Check if the next position is valid
  double nextPosDist = (nextPos - position).getLength();
  if( fabs(nextPosDist - distance) > distanceBuffer 
      || (nextPos - ws.getAgent().getPosition()).getLength() > 1.0 )
  {
    // The next position is invalid, reposition agent.
    //      cout << "MoveSkilling." << endl;
      
    // Calculate the optimum position.
    TVector nMarkDir = markLine.getVector().normalized();
    TVector optMarkPos = nMarkDir * distance + position;

    // If possible, go to the position.
    moveSkill.setDistance(0.5);
    moveSkill.setPosition(optMarkPos);
    if(moveSkill.applicable())
    {
      return moveSkill.generatePlan();
    }
//      else
//      {
//        cout << "Ouch, moveSkill not applicable." << endl;
//      }
  }

  // Select direction to turn to.
  // Select +90 if the ballspeed is to the left of the markLine
  // Select -90 if it's to the right.

  static float turnDir = 90; // Just to avoid jumpy behaviour.

  if( ballIsDistantPosition && ws.getBall().getSpeed().getLength() > 0.4 )
  {
    TVector onDir = markLine.getVector();
    onDir.rotate(90);
    if( (ws.getBall().getSpeed() * onDir) > 0)
      turnDir = 90;
    else
      turnDir = -90;
  }

  // Now the position should be fairly accurate.
  // If neccessary do a turn to adjust the body.
  TAngle bodyDiff = 
    markLine.getVector().getAngle() + turnDir
    - ws.getAgent().getDirection();

  if( fabs(bodyDiff) > 2 )
  {

    // FIXME: Select from +- 90
    // Using ballSpeed?
    //      cout << "Adjusting bodyDir." << endl;

    ws.doActualTurn(bodyDiff);

    // Find the next lookDirection.
    if(lastLook == LK_Distant
       && lookAtMark)
    {
      ws.doTurnNeckTowards(position);
      lastLook = LK_Mark;
    }
    else
    {
      if(ballIsDistantPosition)
	ws.doTurnNeckTowardsBall();
      else
	ws.doTurnNeckTowards(distantPosition);
      lastLook = LK_Distant;
    }

    Plan p = Plan(ws.getStep());
    //      ws.step();
    //        cout << "BodyTurning." << endl;
    //        cout << "New neck dir: " << (ws.getAgent().getNeckDirection()
    //    				   + ws.getAgent().getDirection()) << endl
    //    	   << "Ball: " << (ws.getBall().getPosition()
    //  			   -ws.getAgent().getPosition()).getAngle() << endl
    //  	   << "Mark: " << (position 
    //  			   -ws.getAgent().getPosition()).getAngle() << endl;
      

    return p;
  }
  
  // Otherwise just adjust the position along the line.
  // Adjust the position along the line.

  double moveDistance = (nextPos - ws.getAgent().getPosition())
    * ws.getAgent().getDirectionVector();

  // Adjust for sliding... (Just approx, it won't match anyway...)
//    moveDistance = moveDistance / 1.66;
  moveDistance = moveDistance / 1.4;

  if(fabs(moveDistance) < 0.2)
  {
    moveDistance = 0.0;
//      if(ws.getAgent().getUniformNumber() == 1)
//      cout << "Moron-kick." << endl;
    ws.doKickToSpeed(TVector(3, 0));
  }
  else
  {
    ws.doDashDistance(moveDistance);
  }

  //  cout << "moveDistance: " << moveDistance << endl;

  // Find the next lookPosition.
  if(lastLook == LK_Distant
     && lookAtMark)
  {
    //      cout << "LK_Mark" << endl;
    ws.doTurnNeckTowards(position);
    lastLook = LK_Mark;
  }
  else
  {
    //      cout << "LK_Distant" << endl;
    if(ballIsDistantPosition)
      ws.doTurnNeckTowardsBall();
    else
      ws.doTurnNeckTowards(distantPosition);
    lastLook = LK_Distant;
  }

  Plan p = Plan(ws.getStep());

  //    ws.step();
  //    cout << "LineDashing." << endl;
  //    cout << "New neck dir: " << (ws.getAgent().getNeckDirection()
  //  			       + ws.getAgent().getDirection()) << endl
  //         << "Ball: " << (ws.getBall().getPosition()
  //  		       -ws.getAgent().getPosition()).getAngle() << endl
  //         << "Mark: " << (position 
  //  		       -ws.getAgent().getPosition()).getAngle() << endl;

  return p;
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool MarkPosition::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value
 */

FuzzyBool MarkPosition::applicable()
{
  // If ball is distant position we need to know where the ball is.
  if(ballIsDistantPosition && 
     worldView->getWorldState().getBall().getC() < 0.3)
    return FALSE;

  return TRUE;
}

RS_END_NAMESPACE
