/**
 * @file Pass.cc
 * Implementation of the Pass skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.18 $
 * $Author: mikbr212 $
 * $Date: 2001/07/31 17:06:24 $
 */

#include <stdio.h>

#include "AngleThingy.h"
#include "BallKickable.h"
#include "Pass.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"
/*
#define c_wAdvance   0.5
#define c_wPrecision 1.0
#define c_wOur       1.0
#define c_wTheir    -1.0
*/
#define c_wAdvance   1.0
#define c_wPrecision 1.0
#define c_wOur       1.0
#define c_wTheir    -1.0

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

const float minAng = -135;
const float maxAng = 135;

Pass::Pass()
  : Skill(),
    parameterView(NULL),
    passTreeView(NULL),
    worldStateView(NULL),
//      territoryControlView(NULL),
    kickBall(TVector())
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(passTreeView, PASS_TREE_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
//    getView(territoryControlView, TERRITORY_CONTROL_VIEW_ID,
//    	  64, 64, 32, NULL, NULL, NULL);
  lastCallCycle = 0;
  lastReturnValue = false;
}

// --- Destructor ---

Pass::~Pass()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(PASS_TREE_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(TERRITORY_CONTROL_VIEW_ID);
}

// --- METHODS ---

void Pass::setVictim(const PlayerState& player)
{
  victim = player;
}

Plan Pass::generatePlan()
{
  // maximum/minimum steps the pass should take.
  const int maxSteps = 20;
  const int minSteps = 4; // Fuzzy...
  
  // Opponent penalty for passes.
  const int actionPenalty = 4;

  // How big percentage of max_speed teammates will run in.
  const float teammateEffect = 0.7;

  // How many cycles should be penelty due to observation lag.
  const int observationPenalty = 2;
  
  const int ourActionPenalty = 6; // How heavily penalized are our players.
  const double destSpeed = 0.7; // Desired speed at destination.


  if(calculatedTime == worldStateView->getWorldState().getGameTime())
  {
    return thePlan;
  }
  calculatedTime = worldStateView->getWorldState().getGameTime();
  possible = false;

  // Build anglethingy.
  AngleThingy angles;

  const AgentState& agent
    = worldStateView->getWorldState().getAgent();
  BallState ball
    = worldStateView->getWorldState().getBall();

  // Use heuristics for desired ballSpeed
  double d = parameterView->SP_ball_decay;

  TVector delta = victim.getPosition() - ball.getPosition();
  double v0 = 
    destSpeed + delta.getLength() * (1-d);
  v0 = max(0.0, min(2.2, v0));
  double v0max = delta.getLength() / ( (pow(d, minSteps) - 1) / (d-1));
  v0 = min(v0max, v0);

  angles.setInitialSpeed(v0);

  // Initialize spans.
  AngleThingy::Span initSpan;
  initSpan.start = -179.9;
  initSpan.end = 179.9;
  angles.addSpan(initSpan);

  // Time for heavy black magic.
  // Split the angles
  float bestDir = 0.0;
  float bestSpan = -1.0;
  AngleThingy::Span endSpan;

  int time;

//    // MATLAB
//    FILE* f = fopen("test.m", "wt");

  for(time = 0; time < maxSteps && angles.getSpans().size() > 0; time++)
  {
    angles.step();

    const list<AngleThingy::Span>& spans = angles.getSpans();
    list<AngleThingy::Span>::const_iterator itSpans;

//      // Find the best "safe" direction.
//      if(spans.size() == 0)
//        break;

    // Calculate the agents action radius.
    float ourActionRadius = 0.0;
    if(time > ourActionPenalty)
      ourActionRadius = (time-ourActionPenalty) 
	* parameterView->SP_player_speed_max
	* teammateEffect;

    ourActionRadius += parameterView->SP_kickable_area;

    TVector delta = victim.getPosition() - ball.getPosition();
    float ballRadius = angles.getBallRadius();
    AngleThingy::Span mySpan;
    bool iceptPossible = false;

    // Check if the player intersects the ball circle.
    if( (ourActionRadius+ballRadius) > delta.getLength()
	&& (delta.getLength() + ourActionRadius) > ballRadius)
    {
	TVector deltaOrth = delta.normalized();
	deltaOrth.rotate(-90);
	TVector leftEdge = delta + deltaOrth * ourActionRadius;
	TVector rightEdge = delta - deltaOrth * ourActionRadius;

	mySpan.start = leftEdge.getAngle();
	mySpan.end = rightEdge.getAngle();
	iceptPossible = true;
    }

    for(itSpans = spans.begin(); itSpans != spans.end(); itSpans++)
    {
      if(iceptPossible)
      {
	AngleThingy::Span tmp = *itSpans;
	if(tmp.start > mySpan.end)
	  break;
	if(tmp.end > mySpan.end)
	  tmp.end = mySpan.end;
	if(tmp.start < mySpan.start)
	  tmp.start = mySpan.start;
	  
	if( tmp.end > mySpan.start
	    && (tmp.end - tmp.start) > bestSpan )
	{
	  bestSpan = tmp.end - tmp.start;
	  bestDir = (tmp.end + tmp.start) / 2.0;
	  possible = true;
	  endSpan = tmp;
	}
      }
    }
  }

  if(endSpan.end > delta.getAngle()
     && endSpan.start < delta.getAngle())
  {
    bestDir = delta.getAngle();
  }
  else if(endSpan.start > delta.getAngle())
  {
    bestDir = endSpan.start + 0.2*(endSpan.end - endSpan.start);
  }
  else
  {
    bestDir = endSpan.start + 0.8*(endSpan.end - endSpan.start);
  }

  if(possible)
  {
#ifdef PASS_DEBUG
    cout << "Pass possible, dir: " << bestDir << ", "
	 << agent.getUniformNumber() << "->" 
	 << victim.getUniformNumber() <<endl;
#endif
  }
  else
    {
#ifdef PASS_DEBUG
    cout << "Pass impossible: "
	 << agent.getUniformNumber() << "->" 
	 << victim.getUniformNumber() <<endl;
#endif
    return Plan();
  }

  TVector kickVector = TVector::TPolarVector(v0, bestDir);
  WorldState ws = worldStateView->getWorldState();

  if(victim.getC() < 0.5)
  {
    // Victim confidence is low.
    if(ws.getBall().getSpeed().getLength() < 0.3)
    {
      // The ball is moving slowly.
      // Look at the pass-victim.
      ws = worldStateView->getWorldState();
      ws.doTurnTowards(victim.getPosition());
      ws.doTurnNeckTowards(victim.getPosition());
    }
    else
    {
      // The ball is moving fast, stop it!
      ws = worldStateView->getWorldState();
      ws.doTurnNeckTowards(victim.getPosition());
      ws.doKickToSpeed(TVector(0,0));
    }
  }
  else
  {
    if(ws.getBall().getSpeed().getLength() < 0.2)
    {
#ifdef PASS_DEBUG
      cout << "Ugh! Crappy finishing kick in pass!!" << endl;
#endif
      ws.doKickToSpeed(kickVector);
      ws.doTurnNeckTowardsBall();
    }
    else
    {
      bool kickSuccess = ws.doKickToSpeed(kickVector);
      if(!kickSuccess)
      {
	// We cannot kick right away, stop the gawdamn ball.
	ws = worldStateView->getWorldState();
	ws.doTurnNeckTowards(victim.getPosition());
	ws.doKickToSpeed(TVector(0,0));
      }
    }
  }

  thePlan= Plan(ws.getStep());
  return thePlan;

//    if(!findPass())
//      return kickBall.generatePlan(TVector(1000000000,0));
  
//    return kickBall.generatePlan();
}


/**
 * Persistent.
 * FALSE
 */
FuzzyBool Pass::persistent()
{
  // Never persistent.
  return FALSE;
}

/**
 * Returns the applicable value
 */
FuzzyBool Pass::applicable()
{
//    if(!findPass())
//      return false;
  static BallKickable kickable;
  if(!kickable())
    return FALSE;
  generatePlan();
  if(possible)
    return TRUE;
  return FALSE;
}

RS_END_NAMESPACE

