/*
 * Ball movement
 * a(t) = Power * Power_rate * n =
 *      = Power * Power_rate * (cos(dir), sin(dir))'
 *
 * dir = dir_kicker + dir_command
 * power_rate = kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * a = 0.35 = x * kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * ungefr: x = 0.35 / kick_power_rate
 *
 * Kickable_margin = 0.7
 * kick_power_rate	: 0.016
 * ball_size		: 0.085
 * player_size		: 0.3
 */

/**
 * @file SuperKick.cc
 * Implementation of the SuperKick skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: andbo928 $
 * $Date: 2001/04/15 23:25:45 $
 */

#include "Line.h"
#include "SuperKick.h"
#include "KickCommand.h"
#include "TurnNeckCommand.h"
#include "KickBallToSpeedVector.h"
#include "TAngle.h"

RS_BEGIN_NAMESPACE

  // #define destDir 0
  //agdir
//  #define BODY_DIST 0.5
//  #define BACK_ANGLE 130
//  #define FRONT_ANGLE 55

#define BODY_DIST 0.5
#define BACK_ANGLE 120
#define FRONT_ANGLE 35

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

SuperKick::SuperKick(float dir)
  : Skill(),
    worldView(NULL),
    destDir(dir)
{
  getView(worldView, WORLD_STATE_VIEW_ID);
}

// --- DESTRUCTOR ---

SuperKick::~SuperKick()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}

// --- METHODS ---
void SuperKick::setDirection(float dir)
{
  destDir = dir;
}

float SuperKick::getDirection()
{
  return destDir;
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

bool SuperKick::goalFound(const WorldState& ws)
{
  if(ws.getBall().getSpeed().getLength() < 2.0)
    return false;

  TAngle dirDiff = ws.getBall().getSpeed().getAngle() - destDir;
  if(fabs(dirDiff) < 10)
    {
      return true;
    }

  return false;
}

TVector SuperKick::getPos(int n)
{
  TVector pos;
  switch(n)
    {
    case 0: // LFront
      pos = TVector(-BODY_DIST / std::tan(TAngle::degToRad(-FRONT_ANGLE)),
		    -BODY_DIST);
      break;
    case 1: // LBack
      pos = TVector(-BODY_DIST / std::tan(TAngle::degToRad(-BACK_ANGLE)),
		    -BODY_DIST);
      break;
    case 2: // RFront 
      pos = TVector(BODY_DIST / std::tan(TAngle::degToRad(FRONT_ANGLE)), 
		    BODY_DIST);
      break;
      
    case 3: // RBack
    default:
      pos = TVector(BODY_DIST / std::tan(TAngle::degToRad(BACK_ANGLE)), 
		    BODY_DIST);
      break;
    }

  //  cout << pos << ", l: " << pos.getLength() << ", a: " << pos.getAngle() << endl;
  pos.rotate(destDir);
  return pos;
}


bool SuperKick::opKickSimple(WorldState& ws, Step& opStep)
{
  bool result = ws.doKickToSpeed(TVector::TPolarVector(2.5, destDir));
  opStep = ws.getStep();
  ws.step();
  return result;
}

bool SuperKick::opKickPos1(WorldState& ws, Step& opStep)
{
  TVector pos = TVector(-BODY_DIST / std::tan(TAngle::degToRad(-FRONT_ANGLE)),
			-BODY_DIST);
  pos.rotate(destDir);
  //  pos.rotate(ws.getAgent().getDirection());

  PlayerState nextAgent = ws.getAgent();
  nextAgent.doNull();
  nextAgent.step();

//    cout << "p1 " << pos + nextAgent.getPosition() << endl;

  bool result = ws.doKickToPosition(pos + nextAgent.getPosition());
  ws.doTurnNeckTo(pos.getAngle());
  opStep = ws.getStep();
  ws.step();

  //  cout << "Got pos   : " << ws.getBall().getPosition() << endl;

  return result;
}

bool SuperKick::opKickPos2(WorldState& ws, Step& opStep)
{
  TVector pos = TVector(-BODY_DIST / std::tan(TAngle::degToRad(-BACK_ANGLE)),
			-BODY_DIST);
  pos.rotate(destDir);

  PlayerState nextAgent = ws.getAgent();
  ws.doTurnNeckTo(pos.getAngle());
  nextAgent.doNull();
  nextAgent.step();

//    cout << "p2 " << pos + nextAgent.getPosition() << endl;

  bool result = ws.doKickToPosition(pos + nextAgent.getPosition());
  opStep = ws.getStep();
  ws.step();

  return result;
}

bool SuperKick::opKickPos3(WorldState& ws, Step& opStep)
{
  TVector pos = TVector(BODY_DIST / std::tan(TAngle::degToRad(FRONT_ANGLE)),
			BODY_DIST);
  pos.rotate(destDir);

  PlayerState nextAgent = ws.getAgent();
  nextAgent.doNull();
  nextAgent.step();

//    cout << "p3 " << pos + nextAgent.getPosition() << endl;

  bool result = ws.doKickToPosition(pos + nextAgent.getPosition());
  ws.doTurnNeckTo(pos.getAngle());
  opStep = ws.getStep();
  ws.step();

  return result;
}

bool SuperKick::opKickPos4(WorldState& ws, Step& opStep)
{
  TVector pos = TVector(BODY_DIST / std::tan(TAngle::degToRad(BACK_ANGLE)),
			BODY_DIST);
  pos.rotate(destDir);

  PlayerState nextAgent = ws.getAgent();
  nextAgent.doNull();
  nextAgent.step();

//    cout << "p4 " << pos + nextAgent.getPosition() << endl;

  bool result = ws.doKickToPosition(pos + nextAgent.getPosition());
  ws.doTurnNeckTo(pos.getAngle());
  opStep = ws.getStep();
  ws.step();

  return result;
}


bool SuperKick::genPlanIter(const WorldState& curWs, 
			    int recDepth,
			    Step& theStep)
{

  if(goalFound(curWs))
    return true;

  if(recDepth <= 0)
    return false;

//    cout << "CurNode(" << recDepth << "): "
//         << curWs.getBall().getPosition()
//         << " Speed: " << curWs.getBall().getSpeed().getLength() << " v "
//         << curWs.getBall().getSpeed().getAngle() << endl;

  WorldState nextWS;
  Step opStep;

  // Treat simple kick special, it's always the end of a branch.

  nextWS = curWs;
  opKickSimple(nextWS, theStep);
  if(goalFound(nextWS))
    {
//        cout << "Goal found, speed(" << nextWS.getBall().getSpeed().getLength()
//  	   << ") Plan: simplekick";
      return true;
    }

  bool allFailed = true;

  // Check the four other kicks.
  nextWS = curWs;
  if(opKickPos1(nextWS, opStep))
    {
      allFailed = false;
      if(genPlanIter(nextWS, recDepth - 1, theStep))
	{
//  	  cout << ", kickPos1";
	  theStep = opStep;
	  return true;
	}
    }

  nextWS = curWs;
  if(opKickPos2(nextWS, opStep))
    {
      allFailed = false;
      if(genPlanIter(nextWS, recDepth - 1, theStep))
	{
//  	  cout << ", kickPos2";
	  theStep = opStep;
	  return true;
	}
    }

  nextWS = curWs;
  if(opKickPos3(nextWS, opStep))
    {
      allFailed = false;
      if(genPlanIter(nextWS, recDepth - 1, theStep))
	{
//  	  cout << ", kickPos3";
	  theStep = opStep;
	  return true;
	}
    }

  nextWS = curWs;
  if(opKickPos4(nextWS, opStep))
    {
      allFailed = false;
      if(genPlanIter(nextWS, recDepth - 1, theStep))
	{
//  	  cout << ", kickPos4";
	  theStep = opStep;
	  return true;
	}
    }

//    if(allFailed)
//      curWs.print();
  
  return false;
}

Plan SuperKick::generatePlan()
{
  WorldState ws = worldView->getWorldState();
  ws.reduceCR(0.3, 4);
  // ID-search.
  Step theStep;
//    opKickPos1(ws, theStep);
//    return Plan(theStep);

#define MAX_DEPTH 6
  for(int i = 0; i < MAX_DEPTH; i++)
    {
      if(genPlanIter(ws, i, theStep))
	{
	  //	  cout << endl;
	  //	  cout << "!!!!!!!!!! Plan found, rec-depth: " << i <<  "!!!!!!!!!!!!!" << endl;
	  Plan p = Plan(theStep);
	  return p;
	}
    }
	  
  return Plan();
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool SuperKick::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value
 */

FuzzyBool SuperKick::applicable()
{
  const WorldState& ws = worldView->getWorldState();
  
  // FIXME: Hack
  TVector delta = ws.getAgent().getPosition() - ws.getBall().getPosition();
//    cout << "Applicable: " << ws.getBall().getPosition() << endl;
//    cout << "Delta.len : " << delta.getLength() << endl;
  if(delta.getLength() < 1)
    return TRUE;

  return FALSE;
}

RS_END_NAMESPACE

