/* -*- Mode: C++ -*-
 * $Id: Wait.cc,v 1.1 2001/04/24 17:10:42 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Wait.cc
 * Unitname: Framework/skill
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: mikbr212 $ $Date: 2001/04/24 17:10:42 $
 *
 *
 * HISTORY
 *
 * $Log: Wait.cc,v $
 * Revision 1.1  2001/04/24 17:10:42  mikbr212
 * added
 *
 *
 */

/**
 * @file Wait.cc
 * Implementation file for the skill class Wait.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.1 $
 * $Author: mikbr212 $
 * $Date: 2001/04/24 17:10:42 $
 */


#include "Wait.h"

#include "Display.h"

RS_BEGIN_NAMESPACE




/**
 * Constructor.
 */

Wait::Wait()
  : Skill(),
    gameView(NULL),
    passTreeView(NULL),
    worldStateView(NULL),
    lookView(NULL),
    startCycle(-1),
    maxCycles(30)
{
  getView(gameView, GAME_VIEW_ID);
  getView(passTreeView, PASS_TREE_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(lookView, LOOK_VIEW_ID);
}

/**
 * Destructor.
 */
Wait::~Wait()
{
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(PASS_TREE_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(LOOK_VIEW_ID);
}

Plan Wait::turnPlan() {
  const BallState& ball(worldStateView->getWorldState().getBall());
  AgentState agent(worldStateView->getWorldState().getAgent());
  
  AngleDeg newBodyAngle;
  if(hasLotOfTime) {
    // turn towards (10,0)
    newBodyAngle = AngleDeg((agent.getPosition() - TVector(10,0)).getAngle() + 180); 
  } else {
    // turn towards ball
    newBodyAngle = AngleDeg((ball.getPosition() - agent.getPosition()).getAngle());
  }

  float angle = lookView->getAngle(newBodyAngle - 90, newBodyAngle + 90).getDeg();
  
  newBodyAngle -= agent.getDirection();
  agent.doActualTurn(newBodyAngle.getDeg());
  agent.doTurnNeckTo(angle);
  return Plan(agent.getStep());
}

void Wait::reset() {
  startCycle = -1;
}

bool Wait::timeToBegin() {
  const unsigned int cycleSpan = 3;

  if(teamValues.size() < cycleSpan * 2)
    return false;
  
  float previousValue = 0;
  float currentValue = 0;
  for(unsigned int i=teamValues.size()-cycleSpan; i<teamValues.size(); i++) {
    previousValue += teamValues[i-cycleSpan];
    currentValue += teamValues[i];
  }

  previousValue /= float(cycleSpan);
  currentValue /= float(cycleSpan);

//    cout << "prev = " << previousValue << endl;
//    cout << "current = " << currentValue << endl;

  if(currentValue >= 0.4 || (currentValue > 0.3 && currentValue < previousValue * 0.95))
    return true;
  
  return false;
}

  

bool Wait::wait() {
  static int lastCallCycle = 0;
  static bool lastReturnValue = false;

  const WorldState& world(worldStateView->getWorldState());

  if(lastCallCycle == world.getGameTime())
    return lastReturnValue;
  lastCallCycle = world.getGameTime();
  lastReturnValue = false;
  
  const AgentState& agent(world.getAgent());
  const std::vector<PlayerState>& op(world.getOpponents());
  const PassTree& pt(passTreeView->getAgentPassTree());

  // is this a fresh series of waits?
  if(startCycle == -1) {
    startCycle = world.getGameTime();
    teamValues.clear();
  }

  // has time elapsed?
  if(startCycle + maxCycles < world.getGameTime()) {
    reset();
    return false;
  }
  
  // ----------------------------
  // Now all call-checks are made
  // ----------------------------
  
  // Check trend of teamvalues
  teamValues.push_back(pt.getTeamValue());
  if(timeToBegin()) {
    reset();
    return false;
  }
  
  hasLotOfTime = true;

  // Check playmode
  std::vector<PlayMode> pms(gameView->getPlayModes(NOW));
  std::vector<PlayMode>::iterator pm;
  for(pm = pms.begin(); pm != pms.end(); pm++) {
    switch(*pm) {
    case PM_OUR_OFFSIDE_KICK:
    case PM_OUR_KICK_IN:
    case PM_OUR_CORNER_KICK:
    case PM_GOAL_KICK:
      lastReturnValue = true;
      return true;
    default:
      break;
    }
  }
  
  hasLotOfTime = false;
  const float clearDist = 8;
  // Check if an opponent is close
  std::vector<const PlayerState>::iterator it;
  for(it = op.begin(); it != op.end(); it++) {
    if((it->getPosition() - agent.getPosition()).getLength() < clearDist) {
      reset();
      return false;
    }
  }

  lastReturnValue = true;
  return true;
}

Plan Wait::generatePlan() {
  return turnPlan();
}

FuzzyBool Wait::persistent()
{
  return wait();
}


/**
 * Checks if preconds are true.
 */
FuzzyBool Wait::applicable() {
  const BallState& ball(worldStateView->getWorldState().getBall());
  const AgentState& agent(worldStateView->getWorldState().getAgent());

  if(wait() && agent.isBallKickable(ball)) {
    if(hasLotOfTime)
      return FuzzyBool(1.0);
    else
      return FuzzyBool(0.5);
  }
  return false;
}


RS_END_NAMESPACE


