/* -*- Mode: C++ -*-
 * $Id: StatisticsView.h,v 1.7 2001/07/29 20:53:57 sigge Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: StatisticsView.h
 * Unitname: Framework/view
 * $Revision: 1.7 $
 * Last modified by $Author: sigge $ $Date: 2001/07/29 20:53:57 $
 *
 *
 * HISTORY
 *
 * $Log: StatisticsView.h,v $
 * Revision 1.7  2001/07/29 20:53:57  sigge
 * Added WorldState::getInterceptors and appropriate methods in StatisticsView to accelerate interception calculations.
 *
 * Revision 1.6  2001/07/29 09:43:55  mikbr212
 * Added cyclesSinceOurKickoff
 *
 * Revision 1.5  2001/07/28 14:40:30  sigge
 * Added more intercept cache methods.
 *
 * Revision 1.4  2001/07/12 17:18:02  sigge
 * Added getInterceptionBallState to Statistics view.
 *
 * Revision 1.3  2001/07/06 10:00:38  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 *
 */

/**
 * @file StatisticsView.h
 * Include file for the view class StatisticsView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.7 $
 * $Author: sigge $
 * $Date: 2001/07/29 20:53:57 $
 */


#ifndef STATISTICSVIEW_H
#define STATISTICSVIEW_H


#include "View.h"

#include "TVector.h"
#include <vector>
#include <queue>
#include "BallState.h"
#include "PlayerState.h"

RS_BEGIN_NAMESPACE

class ParameterView;
class WorldStateView;
class GameView;


/**
 * @class StatisticsView StatisticsView.h
 * @par Unit:
 *      Framework/view
 * $Revision: 1.7 $
 * $Author: sigge $
 * $Date: 2001/07/29 20:53:57 $
 */


class StatisticsView : public View {
public:
  enum wsTeam {noTeam, ourTeam, theirTeam};

  StatisticsView(const ViewId id);

  ~StatisticsView();

  // kicker methods
  wsTeam getLastKickerTeam() const;
  int getLastKickerUniformNumber() const;
  int getLastKickCycle() const;

  // opponent kick methods
  TVector getLastOpKickPosition() const;
  TVector getLastOpGoalPosition() const;
  
  // possession methods
  float getPossession() const;

  // player count methods
  int getTeammatesAboveAgent() const;
  int getTeammatesBesideAgent() const;
  int getTeammatesBelowAgent() const;

  int getOpponentsAboveAgent() const;
  int getOpponentsBesideAgent() const;
  int getOpponentsBelowAgent() const;

  int getTeammatesAboveBall() const;
  int getTeammatesBesideBall() const;
  int getTeammatesBelowBall() const;

  int getOpponentsAboveBall() const;
  int getOpponentsBesideBall() const;
  int getOpponentsBelowBall() const;


  // Interceptor stuff
  // Return all interceptor info
  void getInterceptors(PlayerState& ourInterceptor,
		       int& ourInterceptionTime,
		       PlayerState& theirInterceptor,
		       int& theirInterceptionTime);
  // Returns earliest interceptor
  const PlayerState& getInterceptor();
  // Returns their earliest interceptor
  const PlayerState& getTheirInterceptor();
  // Returns our earliest interceptor
  const PlayerState& getOurInterceptor();
  // Returns time before their first intercept
  int getTheirInterceptionTime();
  // Returns time before our first intercept
  int getOurInterceptionTime();
  // Returns time before first intercept
  int getInterceptionTime();
  // Returns predicted ballstate at first intercept
  const BallState& getInterceptionBallState();

  // ball distance methods
  float getOpponentBallDistance() const;
  float getTeammateBallDistance() const;
  bool isClosestToBall(const PlayerState& plr);

  // Game mode methods
  int getCyclesSinceOurKickoff() const;
  int getCyclesSinceTheirKickoff() const;

protected:
  // kicker (and opponent kick) methods
  void clearLastKicker();
  void checkBallKick();

  // possession methods
  void updatePossesion();

  // player count methods
  void updatePlayerCount();
  
  // ball distance methods
  void updateBallDistance();

  // Game mode methods
  void updateCyclesSinceKickoff();
  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();

  void updateInterceptorsCache();

  ParameterView* parameterView;
  WorldStateView* worldStateView;
  GameView* gameView;

  // kicker attributes
  wsTeam lastKickerTeam;
  int lastKickerUniformNumber;
  int lastKickCycle;

  // opponent kick attributes
  TVector lastOpKickPosition;
  TVector lastOpGoalPosition;

  // possesion attributes
  std::queue<int> possessionHist;
  int possessionSum;
  unsigned int maxPossessionLength;

  // player count attributes
  int tmAboveAgent;
  int tmBesideAgent;
  int tmBelowAgent;
  int opAboveAgent;
  int opBesideAgent;
  int opBelowAgent;

  int tmAboveBall;
  int tmBesideBall;
  int tmBelowBall;
  int opAboveBall;
  int opBesideBall;
  int opBelowBall;

  // ball distance attributes
  float opponentBallDistance;
  float teammateBallDistance;

  // Cache
  int interceptionBallStateCacheTime;
  int interceptorsCacheTime;
  PlayerState theirInterceptorCache;
  PlayerState ourInterceptorCache;
  int theirInterceptionTimeCache;
  int ourInterceptionTimeCache;
  BallState interceptionBallState;
  
  // game mode attributes
  int cyclesSinceOurKickoff;
  int cyclesSinceTheirKickoff;

private:
  // Hide copy-constructor and assignment.
  StatisticsView(const StatisticsView&);
  StatisticsView& operator= (const StatisticsView&);  
};


RS_END_NAMESPACE

/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "StatisticsView.icc"
#endif

#endif // !STATISTICSVIEW_H
